/*
 *  Author:	Brian J. Murrell  <spf@interlinx.bc.ca >
 *
 *  Stateful Packet Filter -- constantly adjust an inbound packet filter
 *  to deal with allowing traffic in based on what was sent out.
 *
 *  RealAudio support added by Marc Boucher  <marc@CAM.ORG >
 *  (based on ip_masq_raudio.c by Nigel Metheringham)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef SPF_VIANAME
#define SPF_VIANAME "" // or for example "eth1"
#endif

#include <fstream>
#include <cstdlib>
#include <cctype>
#include <string>
#include <iterator>
#include <strstream>
#include <list.h>
#include <algo.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#if 0
#include <netinet/ip_tcp.h>
#endif
#include <arpa/inet.h>
#include <limits.h>
#include "kernel_headers.h"
#include "icq.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#if !defined(__GLIBC__)  || (__GLIBC__ < 2) || (__GLIBC_MINOR__ < 1)
// I believe that glibc 2.1 mandates that "thou shalt not include O/S headers"
// glibc 2.1 will provide whatever interface is required to access features of
// them, so only include these headers if glibc is less than 2.1
#include <linux/types.h>
#endif

#ifndef IP_OFFMASK	/* not all systems have this defined */
#define IP_OFFMASK 0x1fff	/* mask for fragmenting bits */
#endif

struct fwhdr {
	__u32 len;
	__u32 mark;
	char interface[IFNAMSIZ];
};

// given two forward iterators, search backwards from first to last and return
// another forward iterator
template <class InputIterator, class T>
InputIterator rfind (InputIterator first, InputIterator last, const T& value) {
	while (first != last && *first != value) --first;
	return first;
}

// this should be largest value a time_t can hold without overflowing
const time_t MAX_TIMEOUT = LONG_MAX / CLK_TCK;
const time_t LONG_TIMEOUT = 0x7FFFFFFFL;
const char* DEFAULT_ACTION = "DENY";
const unsigned int TRACEROUTE_MIN = 33435;
const unsigned int TRACEROUTE_MAX = TRACEROUTE_MIN + 30 * 3;
// what rule to start inserting dyn. rules
#if defined(ALLOW_EST_TCP) && !defined(ALLOW_EST_TCP_NORULE)
const unsigned int START_DYN_RULES = 2;
#else
const unsigned int START_DYN_RULES = 1;
#endif

static time_t next_timeout;	// absolute time of next select loop process

// base class for all connection data classes to derive from
class ConnData {
public:
	// construction and destruction
	ConnData () : proto_id (0) { }
	ConnData (const unsigned int p) : proto_id (p) { }
	virtual ~ConnData () { }
	
	// functions to read and write the protocol specific data to a stream
	// each derived class should override these for it's specific data items
	virtual void ostream_data (ostream& o) const { o << proto_id; }
	virtual void ostream_print_data (ostream& o) const { o << " unknown { ";
															 ostream_data (o);
															 o << " }"; }
	virtual void istream_data (istream& i) { }
	
	// data members
	unsigned int proto_id;	// an identifier assigned to this protocol
};

/* remember open icq connections */
class ICQData : public ConnData {
public:
	// construction and destruction
	ICQData () : ConnData (1), lport (0) { }
	ICQData (const unsigned int p) : ConnData (1), lport (p) { }
	~ICQData () { }

	// functions to read and write the protocol specific data to a stream
	void ostream_data (ostream& o) const { o << proto_id << " " << lport; }
	void ostream_print_data (ostream& o) const { o << " icq { ";
												ostream_data (o); o << " }"; }
	void istream_data (istream& i) { i >> lport; }
	
	// data members
	unsigned int lport;
};

class connection {
public:
	struct in_addr src_addr;
	struct in_addr src_mask;
	struct in_addr dst_addr;
	struct in_addr dst_mask;
	uint proto;
	uint src_port_lo;
	uint src_port_hi;
	uint dst_port_lo;
	uint dst_port_hi;
	string interface;
	time_t timeout;
	const ConnData* data;		// store connection data across packets
	connection () { }
	connection (const struct in_addr& sa, const struct in_addr& sm,
			   const struct in_addr& da, const struct in_addr& dm,
			   const uint p, const uint spl, const uint sph, const uint dpl,
			   const uint dph, const string& i, const ConnData* d = 0) :
		src_addr (sa), src_mask (sm), dst_addr (da), dst_mask (dm), proto (p),
		src_port_lo (spl), src_port_hi (sph), dst_port_lo (dpl), dst_port_hi (dph),
		interface (i), timeout (0), data (d) { }
	connection (const struct in_addr& sa, const struct in_addr& sm,
			   const struct in_addr& da, const struct in_addr& dm,
			   const uint p, const uint spl, const uint sph, const uint dpl,
			   const uint dph, const string& i, const time_t t,
			   const ConnData* d = 0) : src_addr (sa), src_mask (sm),
		dst_addr (da), dst_mask (dm), proto (p), src_port_lo (spl),
		src_port_hi (sph), dst_port_lo (dpl), dst_port_hi (dph), interface (i),
		timeout (t), data (d) { }
	~connection () { /* don't do this!!  if (data) delete data; */ }
	bool operator == (const connection& rhs) const {
		if (this == &rhs ||
			// note that timeout is not considered for equality
			(memcmp (&rhs.src_addr, &src_addr, sizeof (src_addr)) == 0 &&
			 memcmp (&rhs.src_mask, &src_mask, sizeof (src_mask)) == 0 &&
			 memcmp (&rhs.dst_addr, &dst_addr, sizeof (dst_addr)) == 0 &&
			 memcmp (&rhs.dst_mask, &dst_mask, sizeof (dst_mask)) == 0 &&
			 rhs.proto == proto &&
			 rhs.src_port_lo == src_port_lo &&
			 rhs.src_port_hi == src_port_hi &&
			 rhs.dst_port_lo == dst_port_lo &&
			 rhs.dst_port_hi == dst_port_hi &&
			 rhs.interface == interface))
			return true;
		return false;
	}
};

list <connection> connections;


static uint debuglev = 0;
// use the preprocessor (which is a no-no in my books) because with GNU stdc++
// you cannot copy iostreams due to ios being declared private
#define debug cerr
static uint asdaemon = 1;
static uint timer_stat = 0;

// timers
#define CLEAR_TIMER(timer) timer.tv_sec = 0; timer.tv_usec = 0;
struct timeval timer_select, timer_setup_select, timer_expire,
	   timer_read_fwdata, timer_do_packet, timer_misc;

static struct in_addr hostmask = { 0xFFFFFFFF };
static struct in_addr zeroaddr = { 0 };
// bjm - we should determine the address for our subnet automagically
// don't forget about yuccky endianess!
static struct in_addr subnetaddr = { 0x000035D1 };
static struct in_addr subnetmask = { 0x00C0FFFF };

typedef list <connection>::iterator SCi;
typedef list <connection>::const_iterator SCci;

extern "C" {
extern uint nsaddrs, ndaddrs;
extern const struct in_addr* saddrs;
extern const struct in_addr* daddrs;
extern int flush_entries (const ip_chainlabel chain);
extern int for_each_chain (int (*fn) (const ip_chainlabel));
extern int delete_chain (const ip_chainlabel chain, int verbose);
extern int delete_entry (const ip_chainlabel chain, const struct ip_fwuser* fw,
						int verbose, int bidir);
extern int new_chain (const ip_chainlabel chain);
extern int ipchains (int argc, char* argv[]);
extern int append_entry (const ip_chainlabel chain, const struct ip_fwuser* fw,
						int verbose, int bidir);
extern int insert_entry (const ip_chainlabel chain, const struct ip_fwuser* fw,
						unsigned int rulenum, int verbose, int bidir);
}

extern void hdump (register const char* , register u_int);
void record_connection (const struct in_addr, const struct in_addr,
					   const struct in_addr, const struct in_addr,
					   const uint, const uint, const uint, const uint,
					   const uint, const time_t, const string&,
					   const ConnData*);
void add_rule (const struct in_addr src, const struct in_addr src_mask,
			  const struct in_addr dst, const struct in_addr dst_mask,
			  const int src_port_lo, const int src_port_hi,
			  const int dst_port_lo, const int dst_port_hi, const uint proto,
			  const time_t secs, const string& interface,
			  const ConnData* data = 0);
inline void add_tcp_pr (const struct in_addr sa, const struct in_addr da,
					   const int spl, const int sph,
					   const int dpl, const int dph,
					   const time_t s, const string& i) {
	add_rule (sa, hostmask, da, hostmask, spl, sph, dpl, dph, IPPROTO_TCP, s, i);
}
inline void add_udp_pr (const struct in_addr sa, const struct in_addr da,
					   const int spl, const int sph,
					   const int dpl, const int dph,
					   const time_t s, const string& i) {
	add_rule (sa, hostmask, da, hostmask, spl, sph, dpl, dph, IPPROTO_UDP, s, i);
}
inline void add_udp_data (const struct in_addr sa, const struct in_addr da,
						 const int sp, const int dp,
						 const time_t s, const string& i, ConnData* d) {
	add_rule (sa, hostmask, da, hostmask, sp, sp, dp, dp, IPPROTO_UDP, s, i, d);
}
inline void add_icmp_r (const struct in_addr sa, const struct in_addr da,
					   const int tl, const int th, const int cl, const int ch,
					   const time_t s, const string& i) {
	add_rule (sa, hostmask, da, hostmask, tl, th, cl, ch, IPPROTO_ICMP, s, i);
}
inline void add_tcp_mask_pr (const struct in_addr sa, const struct in_addr sm,
						 const struct in_addr da, const struct in_addr dm,
						 const int spl, const int sph,
						 const int dpl, const int dph, const time_t s,
						 const string& i) {
	add_rule (sa, sm, da, dm, spl, sph, dpl, dph, IPPROTO_TCP, s, i);
}
inline void add_tcp_mask (const struct in_addr sa, const struct in_addr sm,
						 const struct in_addr da, const struct in_addr dm,
						 const int sp, const int dp, const time_t s,
						 const string& i) {
	add_rule (sa, sm, da, dm, sp, sp, dp, dp, IPPROTO_TCP, s, i);
}
inline void add_udp_mask (const struct in_addr sa, const struct in_addr sm,
						 const struct in_addr da, const struct in_addr dm,
						 const int sp, const int dp, const time_t s,
						 const string& i) {
	add_rule (sa, sm, da, dm, sp, sp, dp, dp, IPPROTO_UDP, s, i);
}
inline void add_icmp_mask (const struct in_addr sa, const struct in_addr sm,
						  const struct in_addr da, const struct in_addr dm,
						  const int t, const int c, const time_t s,
						  const string& i) {
	add_rule (sa, sm, da, dm, t, t, c, c, IPPROTO_ICMP, s, i);
}
inline void add_tcp (const struct in_addr sa, const struct in_addr da,
					const int sp, const int dp, const time_t s,
					const string& i) {
	add_rule (sa, hostmask, da, hostmask, sp, sp, dp, dp, IPPROTO_TCP, s, i);
}
inline void add_udp (const struct in_addr sa, const struct in_addr da,
					const int sp, const int dp, const time_t s,
					const string& i) {
	add_rule (sa, hostmask, da, hostmask, sp, sp, dp, dp, IPPROTO_UDP, s, i);
}
inline void add_icmp (const struct in_addr sa, const struct in_addr da,
					 const int t, const int c, const time_t s,
					 const string& i) {
	add_rule (sa, hostmask, da, hostmask, t, t, c, c, IPPROTO_ICMP, s, i);
}

void cleanup_exit (int);
void catch_signal (int);
void cleanup_chains ();
void set_timer (struct timeval* timer);
char* timer2str (struct timeval* timer);
void install_spf_filters (void);
void add_listener (const struct in_addr src, const struct in_addr src_mask,
				  const struct in_addr dst, const struct in_addr dst_mask,
				  const int src_port_lo, const int src_port_hi,
				  const int dst_port_lo, const int dst_port_hi,
				  const uint proto, const uint cap_size, const string interface,
				  const void *data);
void expire_connections (void);
void do_packet (const char* p, int num_read);
void display_connections (void);
void save_connections (void);
void load_connections (void);
void do_tcp (const char* p, int len);
void do_udp (const char* p);
void do_icmp (const char* p);
void do_icq (const char* p);
void delete_connection (SCci conn);
int mask2bits (struct in_addr in);
void delete_rule (connection& conn);
int atoi (string& s);

int
main (int argc, char* argv[]) {

	char packet[65555];
	int stream, c;
	ssize_t num_read;
	long sec, usec;

	openlog ("spf", LOG_CONS|LOG_NDELAY|LOG_PID, LOG_AUTHPRIV);
	while ((c = getopt (argc, argv, "sx:")) != EOF)
		switch (c) {
			case 'x':
				debuglev = atoi (optarg);
				asdaemon = 0;
				break;
			case 's':
				asdaemon = 0;
				break;
			default:
				cleanup_exit (SIGTERM);
		}
	if ((stream = open ("/dev/fwdata", O_RDONLY)) <0) {
		perror ("open /dev/fwdata");
		exit (1);
	}

	// dump timer stats on a SIGHUP
	signal (SIGHUP, catch_signal);
	// adjust debug level on USR1 and USR2
	signal (SIGUSR1, catch_signal);
	signal (SIGUSR2, catch_signal);

	// cleanup and exit when we get various signals
	signal (SIGINT, catch_signal);
	signal (SIGTERM, catch_signal);
#if 0 // #if 0 this for debugging only!
	signal (SIGSEGV, catch_signal);
#endif
	signal (SIGPIPE, catch_signal);	// in case debug is being piped

	// install filters needed to do stateful filtering
	install_spf_filters ();

	// set the "next timout" to as far into the future as possible
	next_timeout = LONG_TIMEOUT;
	// load any saved rules
	load_connections ();

	if (asdaemon)
		// should go to / so as not to camp on mount points
		chdir ("/");

		// now fork and exit
		if ((c = fork ()) <0) {	// fork failed
			perror ("fork");
			exit (1);
		} else if (c == 0)	// child
			setsid ();
		else			// parent
			exit (0);

	// clear the timers
	CLEAR_TIMER (timer_select);
	CLEAR_TIMER (timer_setup_select);
	CLEAR_TIMER (timer_expire);
	CLEAR_TIMER (timer_read_fwdata);
	CLEAR_TIMER (timer_do_packet);
	CLEAR_TIMER (timer_misc);
	set_timer (NULL);	// reset the timer
	while (1) {
		int ret;
		fd_set readfds;
		struct timeval timeout;

		FD_ZERO (&readfds);
		FD_SET (stream, &readfds);
		// (desperately) expire connections until all expired entries
		// have been purged
		while (next_timeout <= time ((time_t)NULL))
			// check the connections table for expired entries
			expire_connections ();
		set_timer (&timer_expire);
		// kernels > 2.1.126 limit the schedule_timeout () parameter
		// (in ticks) to a signed long.
		timeout.tv_usec = 0;
		timeout.tv_sec = next_timeout-time ((time_t)NULL);
		timeout.tv_sec = min (MAX_TIMEOUT, timeout.tv_sec);
		set_timer (&timer_setup_select);
		if ((ret = select (stream + 1, &readfds, NULL, NULL, &timeout)) <0) {
			set_timer (&timer_select);
			if (errno != EINTR) {
				perror ("select");
				if (debuglev > 0)
					debug << "\nselect (stream+1, &readfds, NULL, NULL, {" <<
			   				 timeout.tv_sec << ", " << timeout.tv_usec << "})";
				cleanup_exit (1);
			}
		} else if (ret == 0) {	// select timeout, go expire
// is there a possibility that we could be so busy to never get here?  If
// there were (say) _always_ data on the netlink device, such that select
// always returned  >  0 we could in theory miss out on expiring connections.
// I have never seen this happen mind you.
// What does select do if you call it with a timeout of 0, yet there are fds
// that need processing?  Well a test reveals that it returns  >  0.  Darn!
			set_timer (&timer_select);
			// check the connections table for expired entries
			expire_connections ();
			set_timer (&timer_expire);
		} else if (ret > 0) {
			set_timer (&timer_select);
			if (FD_ISSET (stream, &readfds)) {
				if ((num_read = read (stream, packet, sizeof (char) * 2000)) <
					0) {
					set_timer (&timer_read_fwdata);
					if (errno == ENOBUFS) {
						perror ("netlink queue overflow (pressing on)");
						continue;
					} else {
						perror ("read");
						cleanup_exit (1);
					}
				}
				set_timer (&timer_read_fwdata);
				do_packet (packet, num_read);
				set_timer (&timer_do_packet);
			}
		}
		if (debuglev > 2)
			debug << '\n';
		if (debuglev > 3)
			display_connections ();
		set_timer (&timer_misc);

		if (timer_stat) {
			timer_stat = 0;
			sec = timer_select.tv_sec + timer_setup_select.tv_sec +
				  timer_expire.tv_sec + timer_read_fwdata.tv_sec +
				  timer_do_packet.tv_sec + timer_misc.tv_sec;
			usec = timer_select.tv_usec + timer_setup_select.tv_usec +
				   timer_expire.tv_usec + timer_read_fwdata.tv_usec +
				   timer_do_packet.tv_usec + timer_misc.tv_usec;
			sec += usec / 1000000;
			usec += usec % 1000000;
			fprintf (stderr, "  select	setup	 expire	 "
							"read   do_packet   misc		TOTAL\n");
			fprintf (stderr, "  (idle)	select	conns	 "
							"fwdata			 timers\n");
			fprintf (stderr, "%s ", timer2str (&timer_select));
			fprintf (stderr, "%s ", timer2str (&timer_setup_select));
			fprintf (stderr, "%s ", timer2str (&timer_expire));
			fprintf (stderr, "%s ", timer2str (&timer_read_fwdata));
			fprintf (stderr, "%s ", timer2str (&timer_do_packet));
			fprintf (stderr, "%s ", timer2str (&timer_misc));
			fprintf (stderr, "%ld.%07ld\n", sec, usec);
			set_timer (&timer_misc);
		}
	}
}

void cleanup_chains () {

	for_each_chain (flush_entries);
	// remove our statinpt chain
	if (delete_chain ("statinpt", 0) < 0)
		perror ("removing statinpt chain");

}

void catch_signal (int sig) {

	switch (sig) {
	case SIGHUP:
		if (debuglev > 0)
			debug << "Got signal to dump stats\n";
		timer_stat = 1;
		break;
	case SIGUSR1:
		if (debuglev < 4)
			debug << "Got signal to increase debug to " << ++debuglev << "\n";
		break;
	case SIGUSR2:
		if (debuglev > 0)
			debug << "Got signal to decrease debug to " << --debuglev << "\n";
		break;
	default:
		if (debuglev > 0)
			debug << "Got signal " << sig << ", cleaning up and exiting\n";
		cleanup_chains ();
		save_connections ();
		if (sig == SIGSEGV)
			abort ();	// leave a core
		closelog ();
		exit (0);
	}

}

// Flush all rules in all chains and exit
void cleanup_exit (int status) {

	if (debuglev > 0)
		debug << "flushing chains and exiting\n";
	cleanup_chains ();
	closelog ();
	exit (status);
}

void do_packet (const char* p, int num_read) {

	char* ip_src,* ip_dst;
	struct ip* ip;

	ip = (struct ip*)(p+sizeof (struct fwhdr));

	if (debuglev > 2) {
		ip_src = strdup (inet_ntoa (ip -> ip_src));
		ip_dst = strdup (inet_ntoa (ip -> ip_dst));
		debug << "sent: (" << num_read << ") " << ip_src << " > " << ip_dst <<
				 " (frag:" << ip -> ip_off << ")";
		if (debuglev > 3)
			hdump (p, num_read);
		free (ip_src);
		free (ip_dst);
	}

	// ignore IP fragments
	if ((ntohs (ip -> ip_off)&IP_OFFMASK) > 0)
		return;
	switch (ip -> ip_p) {
		case IPPROTO_TCP:
			do_tcp (p, num_read - sizeof (struct fwhdr));
			break;
		case IPPROTO_UDP:
			do_udp (p);
			break;
		case IPPROTO_ICMP:
			do_icmp (p);
			break;
		default:
			cerr << "unknown IP protocol \"" << ip -> ip_p << "\"\n";
	}

}

void do_tcp (const char* p, int len) {

	struct ip* ip;
	struct tcphdr* tcp;
	struct fwhdr* fw = (struct fwhdr*)p;
	const char* data;
	const char* data_limit;
	unsigned int data_len;

	if (debuglev > 2)
		debug << " tcp";

	ip = (struct ip*)(p + sizeof (struct fwhdr));
	tcp = (struct tcphdr*)((char*)ip + ip -> ip_hl * 4);
	if (debuglev > 2)
		debug << ' ' << ntohs (tcp -> source) << ' ' << ntohs (tcp -> dest);

#ifndef ALLOW_EST_TCP
	//
	// Allow return TCP packets
	//
	// install a reverse rule for 30 seconds
	add_tcp (ip -> ip_dst, ip -> ip_src, ntohs (tcp -> dest),
		ntohs (tcp -> source), 30, fw -> interface);
#endif
	
	data = (char*)tcp + tcp -> doff * 4;
	data_limit = p + len;

	// length of data is len (the amount captured)
	// minus the length of the IP and TCP headers
	data_len = len - ((ip -> ip_hl * 4) + (tcp -> doff * 4));

	// Watch for and open holes for FTP DATA sessions
	if (ntohs (tcp -> dest) == 21) {
		if (data_len && debuglev > 1) {
			string buf (data, data_len - 2);
			debug << "\nFTP control: \"" << buf << "\"";
		}
		if (debuglev > 3)
			hdump (data, ntohs (ip -> ip_len) - (ip -> ip_hl * 4) -
															(tcp -> doff * 4));
		// this is purely* cheating* cheapo code.  assuming this stuff
		// will always be at byte 0 of a data packet is naive.  the
		// proper solution is to keep a running buffer which we always
		// search for what we want
		if (strncasecmp (data, "port ", 4) == 0) {
			// parse the port specification

			unsigned short int port;

			if (data_len < 20)	// optimize
				return;
			string buf (data, data_len - 2);
			if (debuglev > 0)
				debug << "\nFTP PORT command \"" << buf << "\"";
			string::iterator last_comma = rfind(buf.end() - 1,
												buf.begin(), ',');
			string::iterator secnd_last_comma = rfind(last_comma - 1,
													  buf.begin(), ',');
			if (*last_comma != ',' || *secnd_last_comma != ',') {
				syslog (LOG_NOTICE, "Invalid PORT specification \"%s\"",
					   buf.c_str ());
				return;
			}

			string a (secnd_last_comma + 1, last_comma);
			string b (last_comma + 1, buf.end ());
			port = (atoi (a) << 8) + atoi (b);
			if (debuglev > 0)
				debug << " = port " << port;
			add_tcp (ip -> ip_dst, ip -> ip_src, 20, port, 30, fw -> interface);
		}
	} else if (ntohs (tcp -> source) == 21) {
		if (data_len && debuglev > 1) {
			string buf (data, data_len - 2);
			debug << "\nFTP control: \"" << buf << "\"";
		}
		if (strncasecmp (data, "227 Entering Passive Mode", 25) == 0) {
			// parse the port specification

#if 1 // old - bad: we don't count on the response being split across packets.
	  //			I am not sure of the best way to deal with this yet.
	  //			See below for an idea.
			unsigned short int port;

			if (data_len < 20)	// optimize
				return;
			string buf (data, data_len - 2);
			if (debuglev > 0)
				debug << "\nPassive response: \"" << buf << "\"";
			string::iterator last_comma = rfind(buf.end() - 1,
												buf.begin(), ',');
			string::iterator close_bracket = rfind(buf.end() - 1,
												   last_comma, ')');
			string::iterator secnd_last_comma = rfind(last_comma - 1,
													  buf.begin(), ',');
			if (*last_comma != ',' || *close_bracket != ')' ||
				*secnd_last_comma != ',') {
				syslog (LOG_NOTICE, "\nInvalid PASV port specification \"%s\"",
					   buf.c_str ());
				return;
			}

			string a (secnd_last_comma + 1, last_comma);
			string b (last_comma + 1, close_bracket);
			port = (atoi (a) << 8) + atoi (b);
			if (debuglev > 0)
				debug << " = port ", port;
			add_tcp_pr (ip -> ip_dst, ip -> ip_src, 0, 65535, port,
					   port, 30, fw -> interface);
#else // new - good, not finished
	  //			I think if we create a "connection" record, and keep adding
	  //			received data until we have received a full line (term-
	  //			inated with the \r\n) and then parse it we will be OK.
			connection* conn;
			int l;

			l = ntohs (ip -> ip_len) - (ip -> ip_hl * 4) - (tcp -> doff * 4) - 1;
			if (l >= len)
				l = len - 1;
			data[l] = '\0'; // strrchr needs null-terminated buf
			if (data[l-1] == '\r' && data[l-2] == '\r') {
				// get the host and port and open a rule
				return;
			}
			conn = record_connection (ip -> ip_src, hostmask, ip -> ip_dst,
									 hostmask, ip -> ip_p, ntohs (tcp -> source),
									 ntohs (tcp -> source), ntohs (tcp -> dest),
									 ntohs (tcp -> dest), -1, fw -> interface);	
			// store what we have seen so far
			(char*)conn -> data = strdup (data);
		} else {
			// see if this is another portion of the above
			if ((conn = find (connections.begin (), connections.end (),
							 connection (ip -> ip_src, hostmask,
										ip -> ip_dst, hostmask, ip -> ip_p,
										ntohs (tcp -> source),
										ntohs (tcp -> source),
										ntohs (tcp -> dest),
										ntohs (tcp -> dest),
										fw -> interface))) == connections.end ())
				return;
			// found one.  see if it's complete and process it
			// don't forget to free the connection and it's data
#endif
		}
	} else if ((ntohs (tcp -> dest) == 554) || (ntohs (tcp -> dest) == 7070)) {
		unsigned short version, msg_id, msg_len, udp_port;

		if (!memcmp (data, "OPTIONS", 7) || !memcmp (data, "DESCRIBE", 8)) {
			// This is an RTSP client (not supported for now)
			return;
		}
		
		if (memcmp (data, "PNA", 3)) {
			return; // not initial protocol packet - ignored
		}

		data += 3;

		version = ntohs (*(unsigned short*)data);
		if (debuglev > 0)
			debug << "\nRealAudio: initial seen - protocol version " << version;
		if (version >= 256) {
			if (debuglev > 0)
				debug << "\nRealAudio: protocol version " << version <<
		   				 " not supported";
		}

		data += 2;

		while (data +4 < data_limit) {
			memcpy (&msg_id, data, 2);
			data += 2;
			memcpy (&msg_len, data, 2);
			data += 2;
			if (ntohs (msg_id) == 0) {
				// The zero tag indicates the end of options
				if (debuglev > 0)
					debug << "\nRealAudio: packet end tag seen";
				return;
			}
			if (debuglev > 0)
				debug << "RealAudio: msg " << ntohs (msg_id) - ntohs (msg_len) <<
						 " byte";
			if (ntohs (msg_id) == 0) {
				// The zero tag indicates the end of options
				return;
			}
			p = data;
			data += ntohs (msg_len);
			if (data > data_limit) {
				if (debuglev > 0)
					debug << "\nRealAudio: Packet too short for data";
				return;
			}
			if ((ntohs (msg_id) == 1) || (ntohs (msg_id) == 7)) {
				// MsgId == 1
				// Audio UDP data port on client
				//
				// MsgId == 7
				// Robust UDP error correction port number on
				// client
				//
				// Since these messages are treated just the
				// same, they are bundled together here....
				memcpy (&udp_port, p, 2);
	
				// Sometimes a server sends a message 7 with
				// a zero UDP port.  Rather than do anything
				// with this, just ignore it!
				if (udp_port == 0)
					continue;
	
				// install a reverse rule for 30 seconds
				add_udp_pr (ip -> ip_dst, ip -> ip_src, 0, 65535,
						   ntohs (udp_port), ntohs (udp_port), 30,
						   fw -> interface);
			}
		}
	}
}

void do_udp (const char* p) {

	struct ip* ip;
	struct udphdr* udp;
	struct fwhdr* fw = (struct fwhdr*)p;
	unsigned int timeout;
	SCci conn;

	if (debuglev > 2)
		debug << " udp";

	timeout = 30;	// default timeout for UDP is 30 seconds
	ip = (struct ip*)(p + sizeof (struct fwhdr));
	udp = (struct udphdr*)((char*)ip + ip -> ip_hl * 4);

	// outbound traceroutes should allow corresponding inbound "ttl timed
	// out" messages to the tracerouting host for 15 seconds; as well,
	// allow ICMP port unreachable from the tracrouted host for 15 secs
	if (ntohs (udp -> dest) >= TRACEROUTE_MIN &&
		ntohs (udp -> dest) <= TRACEROUTE_MAX) {

		add_icmp_mask (zeroaddr, zeroaddr, ip -> ip_src, hostmask, 11, 0, 15,
					  fw -> interface);
		add_icmp (ip -> ip_dst, ip -> ip_src, 3, 3, 15, fw -> interface);
		return;	// don't need to allow the UDPs back in
	} else if (ntohs (udp -> dest) == 53) {
		timeout = 90;	// timeout for DNS is 90 seconds
	} else if (ntohs (udp -> dest) == 4000) {
//		timeout = 180;	// timeout for ICQ is 180 seconds
		do_icq (p);
		return;
	} else if ((ntohs (udp -> source) == 68) && (ntohs (udp -> dest) == 67)) {
		// DHCP request
		// we don't know where the response is going to come from
		// although I suppose we could limit to our broadcast domain
		// (our subnet)
		timeout = 10;	// timeout for DHCP is 10 seconds!
		add_udp_mask (subnetaddr, subnetmask, ip -> ip_src, hostmask,
					 ntohs (udp -> dest), ntohs (udp -> source), 10,
					 fw -> interface);
		add_udp_mask (subnetaddr, subnetmask, hostmask, hostmask,
					 ntohs (udp -> dest), ntohs (udp -> source), 10,
					 fw -> interface);
		return;
	}
	//
	// Allow return UDP packets
	//
	
	if (debuglev > 2)
		debug << ' ' << ntohs (udp -> source) << ' ' << ntohs (udp -> dest);
#if 0 // not sure what the "new" code does with respect to "conn"
	// install a reverse rule for timeout seconds
	add_udp (ip -> ip_dst, ip -> ip_src, ntohs (udp -> dest),
			ntohs (udp -> source), timeout, fw -> interface);
#endif
	// check for previously allowed incoming connection first
	conn = find (connections.begin (), connections.end (),
				connection (ip -> ip_src, hostmask, ip -> ip_dst, hostmask,
				IPPROTO_UDP, ntohs (udp -> source), ntohs (udp -> source),
				ntohs (udp -> dest), ntohs (udp -> dest), ""));
	if (conn == connections.end ()) {
		conn = find (connections.begin (), connections.end (),
					connection (ip -> ip_src, hostmask, ip -> ip_dst, hostmask,
					IPPROTO_UDP, 0, 65535, ntohs (udp -> dest),
					ntohs (udp -> dest), ""));
	}
	if (conn != connections.end ()) {
		// install a new rule for timeout seconds
		add_udp (ip -> ip_src, ip -> ip_dst, ntohs (udp -> source),
				ntohs (udp -> dest), timeout, conn -> interface);
	} else {
		// install a reverse rule for timeout seconds
		add_udp (ip -> ip_dst, ip -> ip_src, ntohs (udp -> dest),
				ntohs (udp -> source), timeout, fw -> interface);
	}

}

void do_icmp (const char* p) {

	struct ip* ip;
	struct icmphdr* icmp;
	struct fwhdr* fw = (struct fwhdr*)p;

	if (debuglev > 2)
		debug << " icmp";

	ip = (struct ip*)(p + sizeof (struct fwhdr));
	icmp = (struct icmphdr*)((char*)ip + ip -> ip_hl * 4);
	//
	// Allow ICMP replies matching requests
	//
	
	if (icmp -> type == 8 && icmp -> code == 0) {
		// install a reverse rule for 30 seconds

		if (debuglev > 2)
			debug << " echo request";
		add_icmp (ip -> ip_dst, ip -> ip_src, 0, 0, 30, fw -> interface);
	} else
		if (debuglev > 2)
			debug << " (" << icmp -> type << ", " << icmp -> code << ')';

}

// allow a rule for a packet from a src to a dest for a time limit
void add_rule (const struct in_addr src, const struct in_addr src_mask,
			  const struct in_addr dst, const struct in_addr dst_mask,
			  const int src_port_lo, const int src_port_hi,
			  const int dst_port_lo, const int dst_port_hi, const uint proto,
			  const time_t secs, const string& interface,
			  const ConnData* data = 0) {

	char* ip_src;
	char* ip_dst;
	SCi conn;
	static struct ip_fwuser fw;
	int ret;

	// update an existing entry if there is one
	if ((conn = find (connections.begin (), connections.end (),
					 connection (src, src_mask, dst, dst_mask, proto,
					 src_port_lo, src_port_hi, dst_port_lo, dst_port_hi,
					 interface))) != connections.end ()) {
		conn -> timeout = secs + time ((time_t)NULL);
#if 0	// this is kinda noisy
		if (debuglev > 0) {
			ip_src = strdup (inet_ntoa (src));
			ip_dst = strdup (inet_ntoa (dst));
			debug ("\nupd conn %s:%d %s:%d %u -> %u %u -> %u %u %s %lu",
				  ip_src, mask2bits (src_mask),
				  ip_dst, mask2bits (dst_mask),
				  src_port_lo, src_port_hi, dst_port_lo,
				  dst_port_hi, proto, interface, secs);
			// don't leak memory!
			free (ip_src);
			free (ip_dst);
		}
#endif
// This does not need to be done as conn -> timeout will NEVER be less than
// next_timeout.  Think about it.  The worst case scencario is that next_timeout
// is the old timeout of the connection we have just updated.  The timer will
// expire on the old value and there will be nothing to expire because we have
// asked for the expire time on this connection to be upped.  Simply a wasted
// expire run is all.  Something we should fix, ideally.
//		next_timeout = MIN (next_timeout, conn -> timeout);
//		if (debuglev > 0)
//			debug << "\nnext expire: " << next_timeout - time (NULL);
//
		return;
	}

	if (debuglev > 0) {
		ip_src = strdup (inet_ntoa (src));
		ip_dst = strdup (inet_ntoa (dst));
		debug << "\nadd rule " << ip_src << ":" << mask2bits (src_mask) <<
				 ' ' << ip_dst << ":" << mask2bits (dst_mask) << ' ' <<
				 src_port_lo << " -> " << src_port_hi << ' ' << dst_port_lo <<
				 " -> " << dst_port_hi << ' ' << proto << ' ' << interface <<
				 ' ' << secs;
		if (data)
			data -> ostream_print_data (debug);
		// don't leak memory!
		free (ip_src);
		free (ip_dst);
	}
	fw.ipfw.fw_src = src;
	fw.ipfw.fw_smsk = src_mask;
	fw.ipfw.fw_dst = dst;
	fw.ipfw.fw_dmsk = dst_mask;
	fw.ipfw.fw_proto = proto;
	fw.ipfw.fw_flg = 0;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg = 0;
	fw.ipfw.fw_spts[0] = src_port_lo;
	fw.ipfw.fw_spts[1] = src_port_hi;
	fw.ipfw.fw_dpts[0] = dst_port_lo;
	fw.ipfw.fw_dpts[1] = dst_port_hi;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_outputsize = 0;
	strcpy (fw.ipfw.fw_vianame, interface.c_str ());
	strcpy (fw.label, "ACCEPT");
	nsaddrs = 1;
	saddrs = &src;
	ndaddrs = 1;
	daddrs = &dst;
	if ((ret = insert_entry ("input", &fw, START_DYN_RULES, 0, 0)) != 0) {
		fprintf (stderr, "error inserting a new input rule (ret = %d) at "
						"%s:%d\n0x%x:%u -> %u -> 0x%x:%u -> %u, proto %u\n",
				ret, __FILE__, __LINE__, src.s_addr, src_port_lo, src_port_hi,
				dst.s_addr, dst_port_lo, dst_port_hi, proto);
	} else {
		record_connection (src, src_mask, dst, dst_mask, proto, src_port_lo,
						  src_port_hi, dst_port_lo, dst_port_hi, secs,
						  interface, data);	
	}
	if (debuglev > 1)
		display_connections ();

}

// remove a rule
void delete_rule (SCi conn) {

	static struct ip_fwuser fw;
	int ret;

	memcpy (&fw.ipfw.fw_src, &conn -> src_addr, sizeof (fw.ipfw.fw_src));
	memcpy (&fw.ipfw.fw_smsk, &conn -> src_mask, sizeof (fw.ipfw.fw_smsk));
	memcpy (&fw.ipfw.fw_dst, &conn -> dst_addr, sizeof (fw.ipfw.fw_dst));
	memcpy (&fw.ipfw.fw_dmsk, &conn -> dst_mask, sizeof (fw.ipfw.fw_dmsk));
	fw.ipfw.fw_proto = conn -> proto;
	fw.ipfw.fw_flg = 0;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg = 0;
	fw.ipfw.fw_spts[0] = conn -> src_port_lo;
	fw.ipfw.fw_spts[1] = conn -> src_port_hi;
	fw.ipfw.fw_dpts[0] = conn -> dst_port_lo;
	fw.ipfw.fw_dpts[1] = conn -> dst_port_hi;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_outputsize = 0;
	strcpy (fw.ipfw.fw_vianame, conn -> interface.c_str ());
	strcpy (fw.label, "ACCEPT");
	nsaddrs = 1;
	saddrs = &conn -> src_addr;
	ndaddrs = 1;
	daddrs = &conn -> dst_addr;
	if ((ret = delete_entry ("input", &fw, 0, 0)) != 0) {
		fprintf (stderr, "error deleting an input rule %s:%d\n",
				__FILE__, __LINE__);
	} else
		connections.erase (conn);

}

void record_connection (struct in_addr src, struct in_addr src_mask,
		  struct in_addr dst, struct in_addr dst_mask,
		  uint proto, uint src_port_lo, uint src_port_hi,
		  uint dst_port_lo, uint dst_port_hi, time_t timeout,
		  const string& interface, const ConnData* data) {

/*
	connection* new_conn;
*/

	if (debuglev > 1) {
		char* ip_src;
		char* ip_dst;

		ip_src = strdup (inet_ntoa (src));
		ip_dst = strdup (inet_ntoa (dst));
		debug << "\nrec conn " << ip_src << ":" << mask2bits (src_mask) <<
				 ' ' << ip_dst << ":" << mask2bits (dst_mask) << ' ' <<
				 src_port_lo << " -> " << src_port_hi << ' ' << dst_port_lo <<
				 " -> " << dst_port_hi << ' ' << proto << ' ' <<
				 ' ' << timeout;
		// don't leak memory!
		free (ip_src);
		free (ip_dst);
	}
	// make timeout absolute
	timeout += time ((time_t)NULL);

	connections.push_back (connection (src, src_mask, dst, dst_mask, proto,
									 src_port_lo, src_port_hi, dst_port_lo,
									 dst_port_hi, interface, timeout, data));
	next_timeout = min (next_timeout, timeout);
	if (debuglev > 0)
		debug << "\nnext expire: " << next_timeout - time (NULL);

}

// remove a connection from the table
void delete_connection (SCi conn){
	connections.erase (conn);
}

void display_connections (void) {

	SCci conn;
	char* ip_src;
	char* ip_dst;
	time_t now = time ((time_t)NULL);

	debug << "\nconnections table:";
	for (conn = connections.begin (); conn != connections.end (); conn++) {
		ip_src = strdup (inet_ntoa (conn -> src_addr));
		ip_dst = strdup (inet_ntoa (conn -> dst_addr));
		debug << "\n" << ip_src << ":" << mask2bits (conn -> src_mask) <<
				 ' ' << ip_dst << ":" << mask2bits (conn -> dst_mask) << ' ' <<
				 conn -> src_port_lo << " -> " << conn -> src_port_hi << ' ' <<
				 conn -> dst_port_lo << " -> " << conn -> dst_port_hi << ' ' <<
				 conn -> proto << ' ' << ' ' << conn -> timeout - now;
		if ((*conn).data)
			(*conn).data -> ostream_print_data (debug);
		// don't leak memory!
		free (ip_src);
		free (ip_dst);

	}
	debug << "\nend of connections table";
	fflush (stdout);

}

// save connections to a table to be reloaded when we start again
void load_connections (void) {

	string ip_src, ip_src_mask;
	string ip_dst, ip_dst_mask, interface;
	unsigned int src_port_lo, src_port_hi, dst_port_lo, dst_port_hi, proto,
				 proto_id;
	time_t now = time ((time_t)NULL), timeout;
	ConnData* data;

	if (debuglev > 0)
		debug << "\nloading saved connections from /var/run/spf-conn";
	ifstream conntable ("/var/run/spf-conn");
	if (!conntable)
		// XXX - assume that if there is an error, it's just not there
		return;
	while (conntable) {
		conntable >> ip_src >> ip_src_mask >> ip_dst >> ip_dst_mask >>
					 src_port_lo >> src_port_hi >>
					 dst_port_lo >> dst_port_hi >> 
					 proto >> interface >> timeout >> proto_id;
		if (!conntable)
			break;

		if (debuglev > 0)
			debug << "\nread " << ip_src << ":" << ip_src_mask <<
					 ' ' << ip_dst << ":" << ip_dst_mask << ' ' <<
					 src_port_lo << " -> " << src_port_hi << ' ' <<
					 dst_port_lo << " -> " << dst_port_hi << ' ' <<
					 proto << ' ' << interface << ' ' << timeout << ' ' <<
					 proto_id;
		if (timeout < now)		// skip this one if it expired while
			continue;			// we were shut down

		struct in_addr src_addr, src_mask, dst_addr, dst_mask;
		src_addr.s_addr = inet_addr (ip_src.c_str ());
		src_mask.s_addr = inet_addr (ip_src_mask.c_str ());
		dst_addr.s_addr = inet_addr (ip_dst.c_str ());
		dst_mask.s_addr = inet_addr (ip_dst_mask.c_str ());

		switch (proto_id) {
			case 1:
#if 0
					data = new ICQData;
					data -> proto_id = proto_id;
					data -> istream_data (conntable);
#endif
#if 0
					ICQData icqdata;
					data = &icqdata;
					icqdata.istream_data (conntable);
					icqdata.ostream_print_data (debug);
#endif
					ICQData* icqdata = new ICQData;
					data = icqdata;
					icqdata -> istream_data (conntable);
					if (debuglev > 0)
						icqdata -> ostream_print_data (debug);

					/* add an interface "listener" for the TCP channel */
#if 0 // Shouldn't need to, as it will be installed as the file is read
			 		add_listener (zeroaddr, zeroaddr, dst_addr, hostmask, 0,
								 65535, icqdata.lport, icqdata.lport,
								 IPPROTO_TCP, 65535, interface, NULL);
#endif
					break;
		}
		add_rule (src_addr, src_mask, dst_addr, dst_mask,
				 src_port_lo, src_port_hi, dst_port_lo, dst_port_hi,
				 proto, timeout - now, interface, data);
	}
	if (debuglev > 0)
		debug << "\ndone loading saved connections from /var/run/spf-conn";
	conntable.close ();
	if (debuglev > 0)
		display_connections ();
//	cleanup_exit (0);

}

void save_connections (void) {

	SCci conn;
	char* ip_src;
	char* ip_src_mask;
	char* ip_dst;
	char* ip_dst_mask;

	ofstream conntable ("/var/run/spf-conn", ios::out, 0600);
	if (!conntable) {
		cerr << "error opening /tmp/connections for write" << endl;
		cleanup_exit (1);
	}
	for (conn = connections.begin (); conn != connections.end (); conn++) {
		ip_src = strdup (inet_ntoa (conn -> src_addr));
		ip_src_mask = strdup (inet_ntoa (conn -> src_mask));
		ip_dst = strdup (inet_ntoa (conn -> dst_addr));
		ip_dst_mask = strdup (inet_ntoa (conn -> dst_mask));
		conntable << ip_src << ' ' << ip_src_mask << ' ' <<
					 ip_dst << ' ' << ip_dst_mask << ' ' <<
					 conn -> src_port_lo << ' ' << conn -> src_port_hi << ' ' <<
					 conn -> dst_port_lo << ' ' << conn -> dst_port_hi << ' ' <<
					 conn -> proto << ' ' << conn -> interface << ' ' <<
					 conn -> timeout << ' '; 
		if (! (*conn).data)
			conntable << "0";
		else
			(*conn).data -> ostream_data (conntable);
		conntable << '\n';
		// don't leak memory!
		free (ip_src);
		free (ip_src_mask);
		free (ip_dst);
		free (ip_dst_mask);
	}
	conntable.close ();
}

void expire_connections (void) {

	SCi conn;
	char* ip_src;
	char* ip_dst;

	next_timeout = LONG_TIMEOUT;
	time_t now = time ((time_t)NULL);
	if (debuglev > 1)
		debug << "\nexpiring connections:";
	for (conn = connections.begin (); conn != connections.end (); conn++) {
		if (conn -> timeout <0) {
			// connections with a -ve timeout are for holding
			// state across packets, ignore them for rules
		} else if (conn -> timeout <= now) {
			if (debuglev > 1) {
				ip_src = strdup (inet_ntoa (conn -> src_addr));
				ip_dst = strdup (inet_ntoa (conn -> dst_addr));
				debug << "\nexpired  " << ip_src << ":" <<
						 mask2bits (conn -> src_mask) << ' ' << ip_dst << ":" <<
						 mask2bits (conn -> dst_mask) << ' ' <<
						 conn -> src_port_lo << " -> " << conn -> src_port_hi <<
						 ' ' << conn -> dst_port_lo << " -> " <<
						 conn -> dst_port_hi << ' ' << conn -> proto << ' ' <<
						 ' ' << conn -> timeout - now;
				// don't leak memory!
				free (ip_src);
				free (ip_dst);
			}
			delete_rule (conn++);
			conn--;	// Whoa!  What a fucking hack!
		} else {
			if (debuglev > 1) {
				ip_src = strdup (inet_ntoa (conn -> src_addr));
				ip_dst = strdup (inet_ntoa (conn -> dst_addr));
				debug << "\nnot expd " << ip_src << ":" <<
						 mask2bits (conn -> src_mask) << ' ' << ip_dst << ":" <<
						 mask2bits (conn -> dst_mask) << ' ' <<
						 conn -> src_port_lo << " -> " << conn -> src_port_hi <<
						 ' ' << conn -> dst_port_lo << " -> " <<
						 conn -> dst_port_hi << ' ' << conn -> proto << ' ' <<
						 ' ' << conn -> timeout - now;
				// don't leak memory!
				free (ip_src);
				free (ip_dst);
			}
			next_timeout = min (next_timeout, conn -> timeout);
		}
	}
	if (next_timeout <= now)
		next_timeout = LONG_TIMEOUT;
	if (debuglev > 1)
		debug << "\ndone expiring connections, next expire: " <<
			  next_timeout - now;

}

/* add a port on the output chain to watch for protocol */
void add_listener (const struct in_addr src, const struct in_addr src_mask,
				  const struct in_addr dst, const struct in_addr dst_mask,
				  const int src_port_lo, const int src_port_hi,
				  const int dst_port_lo, const int dst_port_hi,
				  const uint proto, const uint cap_size, const string interface,
				  const void *data) {

	static struct ip_fwuser fw;
	int ret;

	if (debuglev > 0) {
		char* ip_src = strdup (inet_ntoa (src));
		char* ip_dst = strdup (inet_ntoa (dst));
		debug << "\nadd lstn " << ip_src << ":" << mask2bits (src_mask) <<
				 ' ' << ip_dst << ":" << mask2bits (dst_mask) << ' ' <<
				 src_port_lo << " -> " << src_port_hi << ' ' << dst_port_lo <<
				 " -> " << dst_port_hi << ' ' << proto << ' ' << cap_size <<
				 ' ' << interface;
		// don't leak memory!
		free (ip_src);
		free (ip_dst);
	}

	/* install netlink filters on the output chain */
	fw.ipfw.fw_src = src;
	fw.ipfw.fw_smsk = src_mask;
	fw.ipfw.fw_dst = dst;
	fw.ipfw.fw_dmsk = dst_mask;
	fw.ipfw.fw_proto = proto;
	fw.ipfw.fw_flg = IP_FW_F_NETLINK;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg = IP_FW_INV_VIA;
	/* don't want to see fragments */
	fw.ipfw.fw_flg |=  IP_FW_F_FRAG;
	fw.ipfw.fw_invflg |=  IP_FW_INV_FRAG;
	fw.ipfw.fw_spts[0] = src_port_lo;
	fw.ipfw.fw_spts[1] = src_port_hi;
	fw.ipfw.fw_dpts[0] = dst_port_lo;
	fw.ipfw.fw_dpts[1] = dst_port_hi;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_outputsize = cap_size;
	strcpy (fw.ipfw.fw_vianame, interface.c_str ());
	strcpy (fw.label, "ACCEPT");
	nsaddrs = 1;
	saddrs = &fw.ipfw.fw_src;
	ndaddrs = 1;
	daddrs = &fw.ipfw.fw_dst;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	} 
}

void install_spf_filters (void) {

	static struct ip_fwuser fw;
#if 0
	int ret, c, line;
	char* v[100], buf[BUFSIZ];
	FILE* rules;
#endif
	int ret;

	// flush input and output chains
	if ((ret = flush_entries ("input")) != 0) {
		fprintf (stderr, "error flushing input rules %s:%d\n",
				__FILE__, __LINE__);
	}
	if ((ret = flush_entries ("output")) != 0) {
		fprintf (stderr, "error flushing output rules %s:%d\n",
				__FILE__, __LINE__);
	}

	// install netlink filters on the output chain
	fw.ipfw.fw_src.s_addr = 0L;
	fw.ipfw.fw_smsk.s_addr = 0L;
	fw.ipfw.fw_dst.s_addr = 0L;
	fw.ipfw.fw_dmsk.s_addr = 0L;
	fw.ipfw.fw_proto = IPPROTO_ICMP;
	fw.ipfw.fw_flg = IP_FW_F_NETLINK;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg = IP_FW_INV_VIA;
	// don't want to see fragments
	fw.ipfw.fw_flg |= IP_FW_F_FRAG;
	fw.ipfw.fw_invflg |= IP_FW_INV_FRAG;
	fw.ipfw.fw_spts[0] = 0;
	fw.ipfw.fw_spts[1] = 0xFFFF;
	fw.ipfw.fw_dpts[0] = 0;
	fw.ipfw.fw_dpts[1] = 0xFFFF;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_outputsize = 64;
	strcpy (fw.ipfw.fw_vianame, "lo");
	strcpy (fw.label, "ACCEPT");
	nsaddrs = 1;
	saddrs = &fw.ipfw.fw_src;
	ndaddrs = 1;
	daddrs = &fw.ipfw.fw_dst;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}

	fw.ipfw.fw_proto = IPPROTO_UDP;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}

#if defined(ALLOW_EST_TCP)
	// only need to look at TCP output packets for select TCP protos
	fw.ipfw.fw_src.s_addr = 0L;
	fw.ipfw.fw_smsk.s_addr = 0L;
	fw.ipfw.fw_dst.s_addr = 0L;
	fw.ipfw.fw_dmsk.s_addr = 0L;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_proto = IPPROTO_TCP;
	fw.ipfw.fw_flg = IP_FW_F_NETLINK;
	fw.ipfw.fw_invflg = IP_FW_INV_VIA;
	// don't want to see fragments
	fw.ipfw.fw_flg |= IP_FW_F_FRAG;
	fw.ipfw.fw_invflg |= IP_FW_INV_FRAG;
	strcpy (fw.ipfw.fw_vianame, "lo");
	strcpy (fw.label, "ACCEPT");
	nsaddrs = 1;
	saddrs = &fw.ipfw.fw_src;
	ndaddrs = 1;
	daddrs = &fw.ipfw.fw_dst;

	// FTP
	fw.ipfw.fw_spts[0] = 21;
	fw.ipfw.fw_spts[1] = 21;
	fw.ipfw.fw_dpts[0] = 0;
	fw.ipfw.fw_dpts[1] = 0xFFFF;
	fw.ipfw.fw_outputsize = 128;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}
	// FTP
	fw.ipfw.fw_spts[0] = 0;
	fw.ipfw.fw_spts[1] = 0xFFFF;
	fw.ipfw.fw_dpts[0] = 21;
	fw.ipfw.fw_dpts[1] = 21;
	fw.ipfw.fw_outputsize = 128;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}
	// RealAudio
	fw.ipfw.fw_spts[0] = 0;
	fw.ipfw.fw_spts[1] = 0xFFFF;
	fw.ipfw.fw_dpts[0] = 554;
	fw.ipfw.fw_dpts[1] = 554;
	fw.ipfw.fw_outputsize = 128;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}
	fw.ipfw.fw_spts[0] = 0;
	fw.ipfw.fw_spts[1] = 0xFFFF;
	fw.ipfw.fw_dpts[0] = 7070;
	fw.ipfw.fw_dpts[1] = 7070;
	fw.ipfw.fw_outputsize = 128;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}
#else
	// look at all output TCP packets
	fw.ipfw.fw_proto = IPPROTO_TCP;
	if ((ret = insert_entry ("output", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}
#endif

#if defined(ALLOW_EST_TCP) && !defined(ALLOW_EST_TCP_NORULE)
	// install a rule for established TCP on the input chain
	fw.ipfw.fw_src.s_addr = 0L;
	fw.ipfw.fw_smsk.s_addr = 0L;
	fw.ipfw.fw_dst.s_addr = 0L;
	fw.ipfw.fw_dmsk.s_addr = 0L;
	fw.ipfw.fw_proto = IPPROTO_TCP;
	fw.ipfw.fw_flg = IP_FW_F_TCPSYN;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg = IP_FW_INV_SYN;
	fw.ipfw.fw_spts[0] = 0;
	fw.ipfw.fw_spts[1] = 0xFFFF;
	fw.ipfw.fw_dpts[0] = 0;
	fw.ipfw.fw_dpts[1] = 0xFFFF;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_outputsize = 0;
	strcpy (fw.ipfw.fw_vianame, SPF_VIANAME);
	strcpy (fw.label, "ACCEPT");
	nsaddrs = 1;
	saddrs = &fw.ipfw.fw_src;
	ndaddrs = 1;
	daddrs = &fw.ipfw.fw_dst;
	if ((ret = insert_entry ("input", &fw, 1, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}
#endif

	//
	// install user defined rules in their own chain
	//

	delete_chain ("statinpt", 0);	// try to remove the chain first
	if ((ret = new_chain ("statinpt")) == -1) {	// create the new chain
		perror ("Creating statinpt chain");
		cleanup_exit (SIGTERM);
	}
	
	// and link it as the last thing to do in the input chain
	fw.ipfw.fw_src.s_addr = 0L;
	fw.ipfw.fw_smsk.s_addr = 0L;
	fw.ipfw.fw_dst.s_addr = 0L;
	fw.ipfw.fw_dmsk.s_addr = 0L;
	fw.ipfw.fw_proto = 0;
	fw.ipfw.fw_flg = 0;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg = 0;
	fw.ipfw.fw_spts[0] = 0;
	fw.ipfw.fw_spts[1] = 0xFFFF;
	fw.ipfw.fw_dpts[0] = 0;
	fw.ipfw.fw_dpts[1] = 0xFFFF;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_outputsize = 0;
	strcpy (fw.ipfw.fw_vianame, SPF_VIANAME);
	strcpy (fw.label, "statinpt");
	nsaddrs = 1;
	saddrs = &fw.ipfw.fw_src;
	ndaddrs = 1;
	daddrs = &fw.ipfw.fw_dst;
	if ((ret = insert_entry ("input", &fw, START_DYN_RULES, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}
	
#if 0 // don't do this anymore - use ipchains to install initial rules
	if ((rules = fopen ("/etc/spf-rules", "r")) == NULL) {
		if (errno == ENOENT)
			return;
		else {
			perror ("Opening rules file");
			cleanup_exit (SIGTERM);
		}
	}
	// read a line at a time, vectorizing args into an array
	c = 0;
	v[c] = "ipchains";
	line = 0;
	while (!feof (rules)){
		
		char* s;
		char* b;

		line++;
		if (fgets (buf, BUFSIZ, rules) == NULL)
			break;
		// strip off comments
		if ((s = strchr (buf, '#')))
			*s = '\0';
		if ((s = strchr (buf, '\n')))
			*s = '\0';	// get rid of \n at the end
		// skip the line if it's blank
		if (buf[0] == '\0')
			continue;
		b = buf;
		while (1) {
			v[++c] = b;
			if ((s = strchr (b, ' ')) == NULL)
				break;
			while (*s == ' ')
				*s++= '\0';
			b = s;
			// change any input chain specifiers to statinpt
			if (strcmp (v[c], "input") == 0)
				if (strcmp (v[c-1], "-A") == 0 ||
					strcmp (v[c-1], "-D") == 0 ||
					strcmp (v[c-1], "-R") == 0 ||
					strcmp (v[c-1], "-I") == 0 ||
					strcmp (v[c-1], "-P") == 0)
					v[c] = "statinpt";
		}
		// check to see if we got anything at all
		if (c == 0)
			continue;	// nope, next line please
#if 0
		// remove the \n from the last arg
		v[c++][strlen (v[c])-1] = '\0';
#else
		c++;
#endif
		// run the installer
		if (ipchains (c, v) != 0) {
			fprintf (stderr, "error installing rule on line %d\n", line);
			cleanup_exit (SIGTERM);
		}
		c = 0;
	}
	fclose (rules);
#else
	ifstream rules ("/etc/spf-rules");
	if (!rules) {
		cerr << "Error opening /etc/spf-rules\n";
		exit (1);
	}
	
	char s[BUFSIZ];
	unsigned int line_num = 0;
	while (rules.getline (s, BUFSIZ)) {
		line_num++;
		string line (s);

		// remove comments
		string::size_type pound = line.find ('#');
		if (pound != string::npos)
			line.replace (pound, string::npos, "");

		// strip leading whitespace
		if (line[0] == ' ' || line[0] == '\t') {
			string::size_type non_ws = line.find_first_of (" \t", 0);
			line.replace (0, non_ws, "");
		}

		// skip empty lines
		if (line == "")
			continue;

		// change any "input" rules to "statinpt"
		string::size_type input = line.find (" input ");
		if (input != string::npos)
			line.replace (input, 7, " statinpt ");

		// put the ipchains command on front
		line = "ipchains " + line;
		if (system ((line + " >/dev/null 2>&1").c_str ()) != 0) {
			cerr << "error loading rule on line " << line_num << '\n';
			cleanup_exit (1);
		}
	}
#endif
	
	// install default deny filter on the input chain
	fw.ipfw.fw_src.s_addr = 0L;
	fw.ipfw.fw_smsk.s_addr = 0L;
	fw.ipfw.fw_dst.s_addr = 0L;
	fw.ipfw.fw_dmsk.s_addr = 0L;
	fw.ipfw.fw_proto = 0;
	fw.ipfw.fw_flg = IP_FW_F_PRN;
	fw.ipfw.fw_tosand = 0xFF;
	fw.ipfw.fw_tosxor = 0x00;
	fw.ipfw.fw_invflg = 0;
	fw.ipfw.fw_spts[0] = 0;
	fw.ipfw.fw_spts[1] = 0xFFFF;
	fw.ipfw.fw_dpts[0] = 0;
	fw.ipfw.fw_dpts[1] = 0xFFFF;
	fw.ipfw.fw_redirpt = 0;
	fw.ipfw.fw_outputsize = 0;
	strcpy (fw.ipfw.fw_vianame, SPF_VIANAME);
	strcpy (fw.label, DEFAULT_ACTION);
	nsaddrs = 1;
	saddrs = &fw.ipfw.fw_src;
	ndaddrs = 1;
	daddrs = &fw.ipfw.fw_dst;
	if ((ret = append_entry ("input", &fw, 0, 0)) != 0) {
		fprintf (stderr, "error inserting netlink rule %s:%d\n",
				__FILE__, __LINE__);
	}

}

// accumulate time since last call to "timer"
void set_timer (struct timeval* timer) {

	// to remember the time when last called
	static struct timeval last;
	struct timeval now;
	struct timezone tz;
	long sec, usec;

	// What time is it?
	gettimeofday (&now, &tz);
	
	// see if this was a call to reset the "last" timer
	if (!timer) {
		last.tv_sec = now.tv_sec;
		last.tv_usec = now.tv_usec;
		return;
	}
	// subtract last time
	sec = (now.tv_sec - last.tv_sec);
	if ((usec = (now.tv_usec-last.tv_usec)) <1) {
		sec--;
		usec += 1000000;
	}
	timer -> tv_sec += sec;
	if ((timer -> tv_usec += usec) > 1000000) {
		timer -> tv_sec++;
		timer -> tv_usec -= 1000000;
	}
	last.tv_sec = now.tv_sec;
	last.tv_usec = now.tv_usec;
		
}

// convert the value of a timer to a char*
char* timer2str (struct timeval* timer) {

	static char s[40];

	sprintf (s, "%ld.%07ld", timer -> tv_sec, timer -> tv_usec);
	return (s);

}

// convert a bit mask to a number of bits
int mask2bits (struct in_addr in) {
	
	uint32_t a = in.s_addr;
	int bits = 0;
	
	for (int i = 0; i < 32; i++) {
		bits += a & 1;
		a >>= 1;
	}

	return bits;
}

// string version of atoi
int atoi (string& s) {
	strstream strstr;
	int i;
	strstr << s;
	strstr >> i;
	return i;
}

void do_icq (const char* p) {

	struct ip* ip;
	struct udphdr* udp;
	struct fwhdr* fw = (struct fwhdr *)p;
	ICQMsg icqm;
	ICQMsg* pICQMsg = &icqm;
	WORD command;
	unsigned int timeout;
	ICQData* icqdata = new ICQData;
	SCci conn;

	if (debuglev > 2)
		debug << " icq";

	ip = (struct ip*)(p + sizeof (struct fwhdr));
	udp = (struct udphdr*)((char*)ip + ip -> ip_hl * 4);

    pICQMsg->pData = (unsigned char *)udp + sizeof (struct udphdr);
    pICQMsg->bDecrypted = FALSE;
	/* we only act on client messages -- this won't last long however */
    pICQMsg->bServerMsg = FALSE;
    pICQMsg->pNextMsg = NULL;
    pICQMsg->pPrevMsg = NULL;
    pICQMsg->nMsgLen = ntohs (udp->len)-sizeof (struct udphdr);
    pICQMsg->pDecryptedData = (unsigned char *)malloc (2047);
    DecryptPacket (pICQMsg);
	command = ICQCommand (pICQMsg);

	if (debuglev > 0) {
		char cmd[256];
		GetCommandName (command, cmd);
		debug << "\n" << ICQUIN (pICQMsg) << ", " << cmd << " on i/f " <<
				 fw -> interface;
	}
	timeout = 150;	/* I am so damn anal about security!! */
	switch (command) {
		case ICQ_LOGIN:
			icqdata -> lport = (* (DWORD *)((pICQMsg->pDecryptedData) + 24 + 4));
 			add_tcp_mask_pr (zeroaddr, zeroaddr, ip->ip_src, hostmask, 1024,
							65535, icqdata -> lport, icqdata -> lport, timeout,
							fw->interface);
			/* add an interface "listener" for the TCP channel */
 			add_listener (zeroaddr, zeroaddr, ip->ip_src, hostmask, 0, 65535,
						 icqdata -> lport, icqdata -> lport, IPPROTO_TCP, 65535,
						 string (fw->interface), NULL);
			// fall through to install the "server:4000 -> me" rule
		case ICQ_INSTANT_MESSAGE:
		case ICQ_KEEP_ALIVE:
			/* is this an already known open connection? */
 			conn = find (connections.begin (), connections.end (), 
						connection (ip -> ip_dst, hostmask,
								   ip -> ip_src, hostmask,
								   IPPROTO_UDP, ntohs (udp -> dest),
								   ntohs (udp -> dest), ntohs (udp -> source),
								   ntohs (udp -> source), fw -> interface));
			if ((conn != connections.end ()) && (*conn).data) {
				// found data
				delete icqdata;
				icqdata = (ICQData*)conn -> data;
				// is it valid?
				if (icqdata -> lport > 0) {
					if (debuglev > 0) {
						debug << "\nfound connection with data ";
						icqdata -> ostream_print_data (debug);
					}
					/* refresh the listener port rule */
					add_tcp_mask_pr (zeroaddr, zeroaddr, ip -> ip_src, hostmask,
									1024, 65535, icqdata -> lport,
									icqdata -> lport, timeout, fw->interface);
				}
			}
			/* refresh the rule for the server */
 			add_udp_data (ip->ip_dst, ip->ip_src, ntohs (udp->dest),
						 ntohs (udp->source), timeout, fw->interface,
						 icqdata);
			break;
		default:
			delete icqdata;
	}
	if (debuglev > 1)
		display_connections ();

	free (pICQMsg->pDecryptedData);
  
}
